@@ -75,18 +75,30 @@ soup_queue_error_cb (gboolean body_started, gpointer user_data)
 	case SOUP_STATUS_READING_RESPONSE:
 	case SOUP_STATUS_SENDING_REQUEST:
 		if (!body_started) {
-			/*
-			 * This can easily happen if we are using the OpenSSL
-			 * out-of-process proxy and we couldn't establish an
-			 * SSL connection.
-			 */
-			if (req->priv->retries >= 3) {
+			if (req->context->uri->protocol == SOUP_PROTOCOL_HTTPS) {
+				/*
+				 * This can happen if the SSL handshake fails
+				 * for some reason (untrustable signatures,
+				 * etc.)
+				 */
+				if (req->priv->retries >= 3) {
+					soup_message_set_error (
+						req,
+						SOUP_ERROR_SSL_FAILED);
+					soup_message_issue_callback (req);
+				} else {
+					req->priv->retries++;
+					soup_message_requeue (req);
+				}
+			} else if (soup_connection_is_new (req->connection)) {
 				soup_message_set_error (
 					req,
 					SOUP_ERROR_CANT_CONNECT);
 				soup_message_issue_callback (req);
 			} else {
-				req->priv->retries++;
+				/* Must have timed out. Try a new connection */
+				soup_connection_release (req->connection);
+				req->connection = NULL;
 				soup_message_requeue (req);
 			}
 		} else {
@@ -547,6 +559,12 @@ proxy_https_connect (SoupContext    *proxy,
 				  proxy_https_connect_cb,
 				  &ret);
 	soup_message_send (connect_msg);
+
+	/*
+	 * Avoid releasing the connection on message free
+	 */
+	connect_msg->connection = NULL;
+
 	soup_message_free (connect_msg);
 
 	return ret;
@@ -803,7 +821,7 @@ soup_queue_message (SoupMessage    *req,
  * soup_queue_shutdown:
  * 
  * Shut down the message queue by calling %soup_message_cancel on all active
- * requests.
+ * requests and then closing all open connections.
  */
 void 
 soup_queue_shutdown (void)
@@ -820,4 +838,6 @@ soup_queue_shutdown (void)
 	req = soup_queue_first_request ();
 	for (; req; req = soup_queue_next_request ())
 		soup_message_cancel (req);
+
+	soup_connection_purge_idle ();
 }
